Require Export basic.
Require Import ZF ZFsum ZFfix ZFnats ZFord ZFrelations.
Import IZF ZFrepl.

(* Extentionality *)
Lemma cst_is_ext : forall X o, ext_fun o (fun _ => X).
do 2 red; reflexivity.
Qed.
Hint Resolve cst_is_ext.

Lemma sum_is_ext : forall o F G,
  ext_fun o F ->
  ext_fun o G ->
  ext_fun o (fun y => sum (F y) (G y)).
do 2 red; intros.
rewrite (H x x'); trivial.
rewrite (H0 x x'); trivial.
reflexivity.
Qed.
Hint Resolve sum_is_ext.

Lemma func_is_ext : forall x X F,
  ext_fun x F ->
  ext_fun x (fun x => func X (F x)).
do 2 red; intros.
apply func_morph; auto.
reflexivity.
Qed.
Hint Resolve func_is_ext.

(* Continuity *)

Lemma cst_cont : forall X o y, lt y o -> X == sup o (fun _ => X).
intros.
apply eq_intro; intros.
 rewrite sup_ax; trivial.
 exists y; trivial.

 rewrite sup_ax in H0; trivial.
 destruct H0; trivial.
Qed. 

Lemma sum_cont : forall o F G,
  ext_fun o F ->
  ext_fun o G ->
  sum (sup o F) (sup o G) == sup o (fun y => sum (F y) (G y)).
intros.
apply eq_intro; intros.
 rewrite sup_ax; auto.
 elim H1 using sum_ind; clear H1; intros.
  rewrite sup_ax in H1; auto.
  destruct H1.
  exists x0; trivial.
  rewrite H2; apply inl_typ; trivial.

  rewrite sup_ax in H1; auto.
  destruct H1.
  exists x; trivial.
  rewrite H2; apply inr_typ; trivial.

 rewrite sup_ax in H1; auto.
 destruct H1.
 apply sum_mono with (F x) (G x); auto.
Qed.

Require Import ZFord_equiv ZFcard.
Section ProductContinuity.

  Hypothesis mu : set.
  (* Rk: we do not need that mu is a limit ordial *)
  Hypothesis mu_strong : stronglyInaccessibleCard mu.

  Let mu_ord : isOrd mu.
destruct mu_strong as ((isO,_),_); trivial.
Qed.

  (* mu is a regular ordinal *)
  Let mu_regular : forall x F,
    lt_card x mu -> (* we only need card(x) < mu  *)
    ext_fun x F ->
    (forall y, y \in x -> lt (F y) mu) -> lt (sup x F) mu.
destruct mu_strong as (_,reg); trivial.
Qed.

Lemma func_cont : forall X F,
  lt_card X mu ->
  increasing F ->
  func X (sup mu F) == sup mu (fun x => func X (F x)).
intros X F Xsmall Fincr.
assert (Fm : forall o, isOrd o -> ext_fun o F).
 do 2 red; intros.
 apply eq_intro; intros.
  apply Fincr with x; auto.
   apply isOrd_inv with o; auto.
   rewrite <- H1; trivial.

   apply isOrd_inv with o; auto.

   rewrite H1; auto with *.

  apply Fincr with x'; auto.
   apply isOrd_inv with o; auto.

   apply isOrd_inv with o; auto.
   rewrite <- H1; trivial.

   rewrite H1; auto with *.
apply eq_intro; intros.
 pose (G := fun n => ZFordcl.least_ord mu (fun y => app z n \in F y)).
 assert (Gm : ext_fun X G).
  do 2 red; intros.
  unfold G, ZFordcl.least_ord.
  apply union_morph.
  apply subset_morph; try reflexivity.
  red; intros.
  rewrite H1.
  split; destruct 1; split; intros; auto.
   rewrite <- H1 in H6; auto.
   rewrite H1 in H6; auto.
 assert (Fprop : forall x, x \in X -> app z x \in F (G x) /\ lt (G x) mu).
  intros.
  apply app_typ with (x:=x) in H; trivial.
  rewrite sup_ax in H; auto.
  destruct H.
  destruct ZFordcl.least_ord1
    with (o:=mu) (P:=fun y => app z x \in F y) (x:=x0); auto.
   intros.
   apply <- isOrd_equiv in H2. 
   rewrite <- (Fm (osucc x1) (isOrd_succ _ H2) x1 x'); auto.
   apply lt_osucc; trivial.

   destruct H3.
   destruct H4.
   split; trivial.
   apply le_lt_trans with x0; auto.
   red in H4; rewrite succ_equiv in H4; auto.
   apply ZFord.isOrd_inv with mu; trivial.
 assert (Fmu := fun x h => proj2 (Fprop x h)).
 assert (Fspec := fun x h => proj1 (Fprop x h)).
 clear Fprop.
 assert (Ford : forall x, x \in X -> isOrd (G x)).
  intros.
  apply isOrd_inv with mu; auto.
 rewrite sup_ax; auto.
 assert (lt (sup X G) mu).
  apply mu_regular; intros; auto.
 exists (sup X G); trivial.
 apply func_narrow with (1:=H); intros.
 apply Fincr with (G x); auto.
  apply isOrd_inv with mu; auto.

 rewrite sup_ax in H; auto.
 destruct H.
 apply func_mono with X (F x); auto.
 reflexivity.
Qed.

End ProductContinuity.


(***************************************************************************)
(* Elementary types: *)

Definition UNIT := succ zero.
Lemma unit_typ : zero \in UNIT.
unfold UNIT.
apply succ_intro1.
reflexivity.
Qed.

Lemma unit_elim : forall p, p \in UNIT -> p == zero.
unfold UNIT; intros.
elim le_case with (1:=H); intros; trivial.
elim empty_ax with p; trivial.
Qed.

Definition BOOL := succ (succ zero).
Definition TRUE := zero.
Definition FALSE := succ zero.

Lemma true_typ : TRUE \in BOOL.
unfold TRUE, BOOL.
apply succ_intro2.
apply succ_intro1.
reflexivity.
Qed.

Lemma false_typ : FALSE \in BOOL.
unfold FALSE, BOOL.
apply succ_intro1.
reflexivity.
Qed.

Lemma bool_elim : forall b, b \in BOOL -> b == TRUE \/ b == FALSE.
unfold BOOL, TRUE, FALSE; intros.
elim le_case with (1:=H); intros; auto.
elim le_case with (1:=H0); intros; auto.
elim empty_ax with (1:=H1).
Qed.



Section NatOrd_theory.

(***************************************************************************)
(* Natural numbers *)

  Definition NATf X := sum UNIT X.

  Lemma sum_ext_natf : forall A, ext_fun A NATf.
do 2 red; intros.
unfold NATf.
rewrite H0; reflexivity.
Qed.

  Instance NATf_mono : Proper (incl_set ==> incl_set) NATf.
do 2 red; intros.
unfold NATf.
apply sum_mono; trivial.
red; trivial.
Qed.
Hint Resolve NATf_mono Fmono_morph.
  Instance NATf_morph : Proper (eq_set ==> eq_set) NATf.
apply Fmono_morph; trivial.
Qed.


Definition ZERO := inl zero.
Definition SUCC x := inr x.

  Lemma NATf_discr : forall n, ~ ZERO == SUCC n.
red; intros.
apply discr_sum in H; trivial.
Qed.

  Lemma SUCC_inj : forall n m, SUCC n == SUCC m -> n == m.
unfold SUCC.
apply inr_inj.
Qed.

  Lemma ZERO_typ_gen : forall X, ZERO \in sum UNIT X.
unfold ZERO; intros.
apply inl_typ.
apply unit_typ.
Qed.

  Lemma SUCC_typ_gen : forall x X, x \in X -> SUCC x \in sum UNIT X.
unfold SUCC; intros.
apply inr_typ; trivial.
Qed.

  Lemma NATf_case : forall X (P:Prop) x,
    (x == ZERO -> P) ->
    (forall n, n \in X -> x == SUCC n -> P) ->
    x \in NATf X -> P.
intros.
unfold NATf in H.
apply sum_ind with (3:=H1); intros.
 apply H.
 apply unit_elim in H2.
 rewrite H2 in H3; trivial.

 apply H0 with y; trivial.
Qed.

Section IterationNat.

  Definition NATi := TI NATf.

  Instance NATi_morph : morph1 NATi.
unfold NATi; do 2 red; intros.
apply TI_morph; auto.
Qed.


  Lemma NATfun_ext : forall x, ext_fun x (fun n => NATf (NATi n)).
do 2 red; intros.
apply NATf_morph.
apply NATi_morph; trivial.
Qed.
Hint Resolve NATfun_ext.

(* Case analysis *)

  Lemma NATi_case : forall (P:set->Prop) o n,
    isOrd o ->
    (forall x x', x \in NATi o -> x == x' -> P x -> P x') ->
    P ZERO ->
    (forall m o', lt o' o -> m \in NATi o' -> P (SUCC m)) ->    
    n \in NATi o -> P n.
intros.
apply TI_elim in H3; auto.
destruct H3.
apply NATf_case with (3:=H4); intros.
 apply H0 with ZERO; trivial.
  apply TI_intro with x; auto.
  apply ZERO_typ_gen.

  symmetry; trivial.

 apply H0 with (SUCC n0); eauto.
 2:symmetry; trivial.
 apply TI_intro with x; auto.
 apply SUCC_typ_gen; auto.
Qed.

Definition NATCASE_rel (fZ:set) (fS:set->set) (n:set) :=
  fun y =>
    (n == ZERO -> y == fZ) /\
    (forall k, n == SUCC k -> y == fS k).

Lemma NATCASE_choice_pred :
  forall o fZ fS n,
  morph1 fS ->
  isOrd o ->
  n \in NATi o ->
  uchoice_pred (NATCASE_rel fZ fS n).
unfold NATCASE_rel; intros o fZ fS n fsm H H0.
split; intros.
 destruct H2; split; intros; rewrite <- H1; auto.

split; intros.
 apply TI_elim in H0; auto.
 destruct H0.
 elim H1 using NATf_case; intros.
  exists fZ; split; intros; auto with *.
  rewrite H2 in H3; apply NATf_discr in H3; contradiction.

  exists (fS n0); split; intros.
   rewrite H4 in H3; apply NATf_discr in H3; contradiction.

   rewrite H3 in H4; apply SUCC_inj in H4; apply fsm; trivial.

 destruct H1; destruct H2.
 apply TI_elim in H0; auto.
 destruct H0.
 elim H5 using NATf_case; intros.
  rewrite H1; trivial; rewrite H2; auto with *.

  rewrite (H3 n0); trivial; rewrite (H4 n0); auto with *.
Qed.

Definition NATCASE (fZ:set) (fS:set->set) (n:set) :=
  uchoice (NATCASE_rel fZ fS n).

Instance NATCASE_morph : Proper
  (eq_set ==> (eq_set ==> eq_set) ==> eq_set ==> eq_set) NATCASE.
unfold NATCASE; do 4 red; intros.
apply uchoice_morph_raw.
red; intros.
unfold NATCASE_rel.
apply and_iff_morphism.
 rewrite H; rewrite H1; rewrite H2; reflexivity.

 apply fa_morph; intro k.
 rewrite H1; rewrite H2; rewrite (H0 k k); auto with *.
Qed.

Lemma NATCASE_def : forall o fZ fS n,
  morph1 fS ->
  isOrd o ->
  n \in NATi o ->
  NATCASE_rel fZ fS n (NATCASE fZ fS n).
intros.
unfold NATCASE.
apply uchoice_def.
apply NATCASE_choice_pred with o; trivial.
Qed.

Lemma NATCASE_ZERO : forall fZ fS,
  morph1 fS ->
  NATCASE fZ fS ZERO == fZ.
intros.
destruct (NATCASE_def (osucc zero) fZ fS ZERO); trivial.
 apply isOrd_succ; apply isOrd_zero.

 unfold NATi; rewrite TI_mono_succ; trivial.
 apply ZERO_typ_gen.
apply H0; reflexivity.
Qed.

Lemma NATCASE_SUCC : forall o fZ fS n,
  morph1 fS ->
  isOrd o ->
  n \in NATi o ->
  NATCASE fZ fS (SUCC n) == fS n.
intros.
destruct (NATCASE_def (osucc o) fZ fS (SUCC n)); trivial.
 apply isOrd_succ; trivial.

 unfold NATi; rewrite TI_mono_succ; trivial.
 apply SUCC_typ_gen; trivial.

apply H3; reflexivity.
Qed.

Lemma NATCASE_typ :
  forall (P:set->set) o fZ fS,
  morph1 P ->
  morph1 fS ->
  isOrd o ->
  fZ \in P ZERO ->
  (forall n, n \in NATi o -> fS n \in P (SUCC n)) ->
  forall n,
  n \in NATi (osucc o) ->
  NATCASE fZ fS n \in P n.
unfold NATCASE; intros.
assert (nc := NATCASE_def _ fZ _ _ H0 (isOrd_succ _ H1) H4).
destruct nc.
unfold NATi in H4; rewrite TI_mono_succ in H4; auto.
elim H4 using NATf_case; intros.
 rewrite H5; trivial; rewrite H7; trivial.

 rewrite (H6 n0); trivial; rewrite H8; auto.
Qed.

(* Fixpoints *)

  Lemma NATi_fix :
    forall (P:set->Prop) o,
    isOrd o ->
    (forall i, isOrd i -> i \incl o ->
     (forall i' m, lt i' i -> m \in NATi i' -> P m) ->
     forall n, n \in NATi i -> P n) ->
    forall n, n \in NATi o -> P n.
intros P o is_ord Prec.
induction is_ord using isOrd_ind; intros; eauto.
apply IHis_ord.
rewrite (@NATi_morph y y'); trivial.
Qed.

End IterationNat.

Hint Resolve NATfun_ext.

Section NatFixpoint.

(* NAT : the least fixpoint *)

  Definition NAT := NATi omega.

  Lemma NAT_continuous : forall F,
    ext_fun omega F ->
    NATf (sup omega F) == sup omega (fun n => NATf (F n)).
intros.
unfold NATf.
rewrite <- sum_cont; auto.
rewrite <- cst_cont.
 reflexivity.
 apply zero_omega.
Qed.

  Lemma NAT_eq : NAT == NATf NAT.
unfold NAT, NATi.
apply eq_intro; intros.
 rewrite <- TI_mono_succ; auto.
 apply TI_incl with omega; auto.

 unfold NATf in H at 1.
 rewrite TI_eq in H; auto.
 rewrite (cst_cont UNIT omega zero) in H; auto.
 rewrite sum_cont in H; auto.
 rewrite sup_ax in H.
 2:do 2 red; intros; apply NATf_morph; eapply NATfun_ext; eauto.
 destruct H.
 rewrite <- TI_mono_succ in H0; auto.
 apply TI_intro with (osucc x); auto.
 apply isOrd_inv with omega; trivial.
Qed.

  Lemma NATi_NAT : forall o,
    isOrd o ->
    NATi o \incl NAT.
induction 1 using isOrd_ind; intros.
 rewrite <- (NATi_morph y y'); auto.

 unfold NATi.
 rewrite TI_eq; auto.
 red; intros.
 rewrite sup_ax in H2; auto.
 destruct H2.
 rewrite NAT_eq.
 apply NATf_mono with (NATi x); auto.
Qed.

  Lemma ZERO_typ : ZERO \in NAT.
rewrite NAT_eq.
apply ZERO_typ_gen.
Qed.

  Lemma SUCC_typ : forall n, n \in NAT -> SUCC n \in NAT.
intros.
rewrite NAT_eq.
apply SUCC_typ_gen; trivial.
Qed.

  Lemma NAT_fix :
    forall (P:set->Prop),
    (forall o, isOrd o ->
     (forall o' m, lt o' o -> m \in NATi o' -> P m) ->
     forall n, n \in NATi o -> P n) ->
    forall n, n \in NAT -> P n.
intros.
apply NATi_fix with (P:=P) (o:=omega); intros; auto.
apply H with i; trivial.
Qed.

  Lemma NAT_ind : forall (P:set->Prop),
    (forall x x', x \in NAT -> x == x' -> P x -> P x') ->
    P ZERO ->
    (forall n, n \in NAT -> P n -> P (SUCC n)) ->    
    forall n, n \in NAT -> P n.
intros.
elim H2 using NAT_fix; intros.
elim H5 using NATi_case; trivial; intros.
 apply H with x; trivial.
 apply NATi_NAT with o; trivial.

 apply H1; eauto.
 apply NATi_NAT with o'; trivial.
 apply isOrd_inv with o; trivial.
Qed.

  Fixpoint nat2NAT (n:nat) : set :=
    match n with
    | 0 => ZERO
    | S k => SUCC (nat2NAT k)
    end.

  Lemma nat2NAT_reflect : forall x,
    x \in NAT -> exists n, x == nat2NAT n.
intros.
elim H using NAT_ind; intros.
 destruct H2.
 exists x1.
 rewrite <- H1; rewrite H2; reflexivity.

 exists 0; reflexivity.

 destruct H1.
 exists (S x0); unfold SUCC; rewrite H1; reflexivity.
Qed.

End NatFixpoint.

Existing Instance NATi_morph.

Section NatUniverse.

  (* The universe where we build the inductive type *)
  Variable U : set.
  Hypothesis has_cstr : forall X, X \in U -> sum UNIT X \in U.
  Hypothesis has_empty : empty \in U.
  Hypothesis has_limit : forall F,
    ext_fun omega F ->
    (forall n, n \in omega -> F n \in U) -> sup omega F \in U.

  Lemma has_fin_limit : forall n F,
    lt n omega ->
    ext_fun n F ->
    (forall m, lt m n -> F m \in U) -> sup n F \in U.
Admitted.

  Lemma NATi_pre_typ : forall n, lt n omega -> NATi n \in U.
intros.
apply isOrd_sup_elim in H.
destruct H.
revert n H.
induction x; simpl in *; intros.
 elim empty_ax with (1:=H).

 unfold NATi; rewrite TI_eq; auto.
 apply has_fin_limit; intros; auto.
  apply isOrd_trans with (2:=H); auto.
  apply isOrd_sup_intro with (S (S x)); simpl.
  apply lt_osucc_compat; auto.

  apply has_cstr.
  apply IHx.
  apply olts_le in H.
  apply H in H0; trivial.

 apply isOrd_inv with (2:=H); auto.
Qed.

  Lemma NAT_typ : NAT \in U.
unfold NAT, NATi.
rewrite TI_eq; auto.
apply has_limit; intros; auto.
apply has_cstr.
apply NATi_pre_typ; trivial.
Qed.

End NatUniverse.


(***************************************************************************)
(* Ordinals *)

  Definition ORDf X := sum UNIT (func NAT X).

  Lemma sum_ext_ordf : forall A, ext_fun A ORDf.
do 2 red; intros.
unfold ORDf.
rewrite H0; reflexivity.
Qed.

  Instance ORDf_mono : Proper (incl_set ==> incl_set) ORDf.
do 2 red; intros.
unfold ORDf.
apply sum_mono; trivial.
red; trivial.
apply func_mono; trivial.
reflexivity.
Qed.
Hint Resolve ORDf_mono.
  Instance ORDf_morph : Proper (eq_set ==> eq_set) ORDf.
apply Fmono_morph; trivial.
Qed.

  Definition LIM f := inr f.

  Lemma LIM_typ_gen : forall x X, x \in func NAT X -> LIM x \in sum UNIT (func NAT X).
unfold LIM; intros.
apply inr_typ; trivial.
Qed.

  Lemma ORDf_case : forall X (P:Prop) x,
    (x == ZERO -> P) ->
    (forall f, f \in func NAT X -> x == LIM f -> P) ->
    x \in ORDf X -> P.
intros.
unfold ORDf in H.
apply sum_ind with (3:=H1); intros.
 apply H.
 apply unit_elim in H2.
 rewrite H2 in H3; trivial.

 apply H0 with y; trivial.
Qed.

Section IterationOrd.

  Definition ORDi := TI ORDf.

  Instance ORDi_morph : morph1 ORDi.
unfold ORDi; do 2 red; intros.
apply TI_morph; trivial.
Qed.

  Lemma ORDfun_ext : forall x, ext_fun x (fun n => ORDf (ORDi n)).
do 2 red; intros.
apply ORDf_morph.
apply ORDi_morph; trivial.
Qed.
Hint Resolve ORDfun_ext.

  Lemma ORDi_case : forall (P:set->Prop) o n,
    isOrd o ->
    (forall x x', x \in ORDi o -> x == x' -> P x -> P x') ->
    P ZERO ->
    (forall f o', lt o' o -> f \in func NAT (ORDi o') -> P (LIM f)) ->    
    n \in ORDi o -> P n.
intros.
apply TI_elim in H3; auto.
destruct H3.
apply ORDf_case with (3:=H4); intros.
 apply H0 with ZERO; trivial.
  apply TI_intro with x; auto.
  apply ZERO_typ_gen.

  symmetry; trivial.

 apply H0 with (LIM f); eauto.
 2:symmetry; trivial.
 apply TI_intro with x; auto.
 apply LIM_typ_gen; auto.
Qed.

  Lemma ORDi_fix :
    forall (P:set->Prop) o,
    isOrd o ->
    (forall i, isOrd i -> i \incl o ->
     (forall i' m, lt i' i -> m \in ORDi i' -> P m) ->
     forall n, n \in ORDi i -> P n) ->
    forall n, n \in ORDi o -> P n.
intros P o is_ord Prec.
induction is_ord using isOrd_ind; intros; eauto.
apply IHis_ord.
rewrite (@ORDi_morph y y'); trivial.
Qed.

End IterationOrd.

Hint Resolve ORDfun_ext.

Section OrdFixpoint.

(* ORD : the least fixpoint *)

  Hypothesis mu : set.
  Hypothesis mu_strong : stronglyInaccessibleCard mu.
  (* mu is a cardinal bigger than those of the types of our universe *)
  Hypothesis NAT_in_mu : lt_card NAT mu.

  Let mu_ord : isOrd mu.
destruct mu_strong as ((isO,_),_); trivial.
Qed.

  (* mu is a limit ordial *)
  Let mu_lim : forall x, lt x mu -> lt (osucc x) mu.
destruct mu_strong as ((_,lim),_); trivial.
Qed.

  Definition ORD := ORDi mu.

  Lemma ORD_eq : ORD == ORDf ORD.
unfold ORD, ORDi.
apply eq_intro; intros.
 rewrite <- TI_mono_succ; auto.
 apply TI_incl with mu; auto.

 unfold ORDf in H at 1.
 rewrite TI_eq in H; auto.
 rewrite (cst_cont UNIT mu zero) in H; auto.
 2:apply lt_card_non_zero in NAT_in_mu; trivial.
 rewrite func_cont in H; auto.
  rewrite sum_cont in H; auto.
  rewrite sup_ax in H.
  2:do 2 red; intros; apply ORDf_morph; eapply ORDfun_ext; eauto.
  destruct H.
  rewrite <- TI_mono_succ in H0; auto.
   apply TI_intro with (osucc x); auto.
   apply isOrd_inv with mu; trivial.

  red; intros.
  apply ORDf_mono.
  apply TI_mono; auto.
Qed.

  Lemma ORDi_ORD : forall o,
    isOrd o ->
    ORDi o \incl ORD.
induction 1 using isOrd_ind; intros.
 rewrite <- (ORDi_morph y y'); auto.

 unfold ORDi.
 rewrite TI_eq; auto.
 red; intros.
 rewrite sup_ax in H2; auto.
 destruct H2.
 rewrite ORD_eq.
 apply ORDf_mono with (ORDi x); auto.
Qed.

Section OrdUniverse.

  (* The universe where we build the inductive type *)
  Variable U : set.
  Hypothesis has_cstr : forall X, X \in U -> sum UNIT (func NAT X) \in U.
  Hypothesis has_empty : empty \in U.
  Hypothesis mu_clos : forall x F,
    le x mu ->
    ext_fun x F ->
    (forall n, n \in x -> F n \in U) -> sup x F \in U.

  Lemma ORDi_pre_typ : forall n, lt n mu -> ORDi n \in U.
intros.
assert (isOrd n).
 apply isOrd_inv with mu; auto.
revert H.
induction H0 using isOrd_ind; intros.
 rewrite <- H1 in H2.
 rewrite <- (ORDi_morph y y'); auto.

 unfold ORDi; rewrite TI_eq; auto.

 apply mu_clos; intros; auto.
 apply has_cstr; apply H1; trivial.
 apply isOrd_trans with y; trivial.
Qed.

  Lemma ORD_typ : ORD \in U.
unfold ORD, ORDi.
rewrite TI_eq; auto.
apply mu_clos; intros; auto.
apply has_cstr.
apply ORDi_pre_typ; trivial.
Qed.

End OrdUniverse.

End OrdFixpoint.

End NatOrd_theory.

Hint Resolve NATf_mono Fmono_morph NATfun_ext.
